interface IReadonlyQueue<T> {
    /**
     * Iterable.
     */
    [Symbol.iterator](): IterableIterator<T>;
    /**
     * Determine whether the data-structure is destroyed.
     */
    readonly destroyed: boolean;
    /**
     * Count the element in the collection.
     * @getter
     */
    readonly size: number;
    /**
     * Count the element in the queue which is accepted by the filter.
     */
    count(filter: (element: T) => boolean): number;
    /**
     * Get the front element of the queue.
     */
    front(): T | undefined;
}
interface IQueue<T> extends IReadonlyQueue<T> {
    /**
     * Destroy the data-structure.
     */
    destroy(): void;
    /**
     * Reset the collection and insert the initial elements.
     */
    init(initialElements?: Iterable<T>): void;
    /**
     * Popup the elements from the queue by the `dequeue` order.
     * Once the queue first empty, the iterator will be ended.
     */
    consuming(): IterableIterator<T>;
    /**
     * Popup the front element from the queue.
     * @param newElement if specified, the element will be enqueued, it is for better performance.
     */
    dequeue(newElement?: T): T | undefined;
    /**
     * Add an element to the tail of the queue.
     */
    enqueue(element: T): void;
    /**
     * Add multiple elements to the tail of the queue.
     */
    enqueues(elements: Iterable<T>): void;
    /**
     * Add multiple elements to the tail of the queue.
     *
     * !!!You should ensure the `start` and `end` are valid index by yourself.!!!
     */
    enqueues_advance(elements: ReadonlyArray<T>, start: number, end: number): void;
    /**
     * Remove existed elements which is accepted by the filter,
     * @returns the number of removed elements.
     */
    exclude(filter: (element: T) => boolean): number;
}

interface IReadonlyDeque<T> extends IReadonlyQueue<T> {
    /**
     * Get the last element of the queue.
     */
    back(): T | undefined;
}
interface IDeque<T> extends IReadonlyDeque<T>, IQueue<T> {
    /**
     * Popup the last element of the queue.
     */
    dequeue_back(): T | undefined;
    /**
     * Add an element to the front of the queue.
     */
    enqueue_front(element: T): void;
    /**
     * Add multiple elements to the front of the queue.
     */
    enqueues_front(elements: Iterable<T>): void;
    /**
     * Add multiple elements to the front of the queue.
     *
     * !!!You should ensure the `start` and `end` are valid index by yourself.!!!
     */
    enqueues_front_advance(elements: ReadonlyArray<T>, start: number, end: number): void;
}

interface IReadonlyCircularQueue<T> extends IReadonlyDeque<T> {
}
interface ICircularQueue<T> extends IReadonlyCircularQueue<T>, IDeque<T> {
    /**
     * Find the element at the given index.
     */
    at(index: number): T | undefined;
    /**
     * Resize the max-size of queue with the given size.
     */
    resize(MAX_SIZE: number): void;
    /**
     * Rearrange elements, that is, put the first element in place 0-index.
     */
    rearrange(): void;
}

interface IFixedCircularQueueProps {
    /**
     * Initial capacity of the circular queue.
     */
    readonly capacity: number;
}
declare class CircularQueue<T = unknown> implements ICircularQueue<T> {
    protected readonly _elements: T[];
    protected _capacity: number;
    protected _size: number;
    protected _start: number;
    protected _end: number;
    protected _destroyed: boolean;
    constructor(props: IFixedCircularQueueProps);
    [Symbol.iterator](): IterableIterator<T>;
    get destroyed(): boolean;
    get size(): number;
    at(index: number): T | undefined;
    count(filter: (element: T) => boolean): number;
    front(): T | undefined;
    back(): T | undefined;
    destroy(): void;
    init(initialElements?: Iterable<T>): void;
    resize(newCapacity: number): void;
    rearrange(): void;
    consuming(): IterableIterator<T>;
    dequeue(newElement?: T): T | undefined;
    enqueue(element: T): void;
    enqueues(elements: Iterable<T>): void;
    enqueues_advance(elements: ReadonlyArray<T>, start: number, end: number): void;
    exclude(filter: (element: T) => boolean): number;
    dequeue_back(): T | undefined;
    enqueue_front(element: T): void;
    enqueues_front(elements: Iterable<T>): void;
    enqueues_front_advance(elements: ReadonlyArray<T>, start: number, end: number): void;
}

interface IDequeNode<T> {
    value: T;
    prev: IDequeNode<T> | undefined;
    next: IDequeNode<T> | undefined;
}
declare class Deque<T = unknown> implements IDeque<T> {
    protected readonly _pool: Array<IDequeNode<T>>;
    protected _poolSize: number;
    protected _size: number;
    protected _head: IDequeNode<T> | undefined;
    protected _tail: IDequeNode<T> | undefined;
    protected _destroyed: boolean;
    constructor();
    [Symbol.iterator](): IterableIterator<T>;
    get destroyed(): boolean;
    get size(): number;
    count(filter: (element: T) => boolean): number;
    front(): T | undefined;
    back(): T | undefined;
    destroy(): void;
    init(initialElements?: Iterable<T>): void;
    consuming(): IterableIterator<T>;
    dequeue(newElement?: T | undefined): T | undefined;
    enqueue(element: T): void;
    enqueues(elements: Iterable<T>): void;
    enqueues_advance(elements: ReadonlyArray<T>, start: number, end: number): void;
    exclude(filter: (element: T) => boolean): number;
    dequeue_back(): T | undefined;
    enqueue_front(element: T): void;
    enqueues_front(elements: Iterable<T>): void;
    enqueues_front_advance(elements: ReadonlyArray<T>, start: number, end: number): void;
}

/**
 * Type for a comparison function.
 *
 * - -1 or other negative number if the first argument is less than the second
 * - 0 if they are equal,
 * - 1 or other positive number if the first argument is greater than the second.
 */
type ICompare<T> = (x: T, y: T) => -1 | 0 | 1 | number;

interface IReadonlyPriorityQueue<T> extends IReadonlyQueue<T> {
}
interface IPriorityQueue<T> extends IReadonlyPriorityQueue<T>, IQueue<T> {
}

interface IPriorityQueueProps<T> {
    /**
     * If the compare(x, y) < 0, then x has a higher priority than y.
     */
    compare: ICompare<T>;
}
/**
 * MinHeap
 */
declare class PriorityQueue<T> implements IPriorityQueue<T> {
    protected readonly _compare: ICompare<T>;
    protected readonly _elements: T[];
    protected _size: number;
    protected _destroyed: boolean;
    constructor(props: IPriorityQueueProps<T>);
    /**
     * !!NOTICE!! The iterator will not be sorted.
     */
    [Symbol.iterator](): IterableIterator<T>;
    get destroyed(): boolean;
    get size(): number;
    count(filter: (element: T) => boolean): number;
    front(): T | undefined;
    destroy(): void;
    init(initialElements?: Iterable<T>): void;
    consuming(): IterableIterator<T>;
    dequeue(element?: T): T | undefined;
    enqueue(element: T): void;
    enqueues(elements: Iterable<T>): void;
    enqueues_advance(elements: ReadonlyArray<T>, start: number, end: number): void;
    exclude(filter: (element: T) => boolean): number;
    protected _down(index: number): void;
    protected _up(index: number): void;
    protected _fastBuild(): void;
}

export { CircularQueue, Deque, type ICircularQueue, type IDeque, type IFixedCircularQueueProps, type IPriorityQueue, type IPriorityQueueProps, type IQueue, type IReadonlyCircularQueue, type IReadonlyDeque, type IReadonlyPriorityQueue, type IReadonlyQueue, PriorityQueue };
